@@ -275,6 +275,27 @@ notify_name_change (AtkObject *atk_obj)
   if (gail_label->textutil == NULL)
     return;
 
+  /*
+   * Check whether the label has actually changed before emitting
+   * notification.
+   */
+  if (gail_label->textutil->buffer) 
+    {
+      GtkTextIter start, end;
+      const char *new_label;
+      char *old_label;
+      int same;   
+
+      gtk_text_buffer_get_start_iter (gail_label->textutil->buffer, &start);
+      gtk_text_buffer_get_end_iter (gail_label->textutil->buffer, &end);
+      old_label = gtk_text_buffer_get_text (gail_label->textutil->buffer, &start, &end, FALSE);
+      new_label = gtk_label_get_text (label);
+      same = strcmp (new_label, old_label);
+      g_free (old_label);
+      if (same == 0)
+        return;
+    }
+   
   /* Create a delete text and an insert text signal */
  
   g_signal_emit_by_name (gail_obj, "text_changed::delete", 0, 
@@ -352,6 +373,7 @@ gail_label_real_notify_gtk (GObject           *obj,
       gint start, end;
       gboolean text_caret_moved = FALSE;
       gboolean selection_changed = FALSE;
+      gboolean is_start = TRUE;
 
       gail_obj = G_OBJECT (atk_obj);
       label = GTK_LABEL (widget);
@@ -361,13 +383,11 @@ gail_label_real_notify_gtk (GObject           *obj,
           if (start != gail_label->cursor_position ||
               end != gail_label->selection_bound)
             {
+              if (end != gail_label->selection_bound)
+                is_start = FALSE;
               gail_label->selection_bound = end;
-              if (start != gail_label->cursor_position)
-                {
-                  /* cursor movement */
-                  gail_label->cursor_position = start;
-                  text_caret_moved = TRUE;
-                }
+              gail_label->cursor_position = start;
+              text_caret_moved = TRUE;
               if (start != end)
                 selection_changed = TRUE;
             }
@@ -378,12 +398,15 @@ gail_label_real_notify_gtk (GObject           *obj,
             selection_changed = TRUE;
           if (gtk_label_get_selectable (label))
             {
-              gint old_cursor_position = gail_label->cursor_position;
-
+              if (gail_label->cursor_position != -1 && start != gail_label->cursor_position)
+                text_caret_moved = TRUE;
+              if (gail_label->selection_bound != -1 && end != gail_label->selection_bound)
+                {
+                  text_caret_moved = TRUE;
+                  is_start = FALSE;
+                }
               gail_label->cursor_position = start;
               gail_label->selection_bound = end;
-              if (old_cursor_position != -1 && start != old_cursor_position)
-                text_caret_moved = TRUE;
             }
           else
             {
@@ -393,11 +416,11 @@ gail_label_real_notify_gtk (GObject           *obj,
               gail_label->selection_bound = 0;
               text_caret_moved = TRUE;
             }
-            
+
         }
         if (text_caret_moved)
           g_signal_emit_by_name (gail_obj, "text_caret_moved", 
-                                 gail_label->cursor_position);
+                                 is_start ? gail_label->cursor_position : gail_label->selection_bound);
         if (selection_changed)
           g_signal_emit_by_name (gail_obj, "text_selection_changed");
 
@@ -467,14 +490,34 @@ gail_label_ref_relation_set (AtkObject *obj)
           AtkObject *accessible_array[1];
           AtkRelation* relation;
 
-          /*
-           * Handle the case where a GnomeIconEntry is specified as the 
-           * mnemonic widget. use the button which is a grandchild of the
-           * GnomeIconEntry as the mnemonic widget. See bug #133967.
-           */
           if (!GTK_WIDGET_CAN_FOCUS (mnemonic_widget))
             {
-              if (GTK_IS_BOX (mnemonic_widget))
+            /*
+             * Handle the case where a GtkFileChooserButton is specified as the 
+             * mnemonic widget. use the combobox which is a child of the
+             * GtkFileChooserButton as the mnemonic widget. See bug #359843.
+             */
+             if (GTK_IS_BOX (mnemonic_widget))
+               {
+                  GList *list, *tmpl;
+
+                  list = gtk_container_get_children (GTK_CONTAINER (mnemonic_widget));
+                  if (g_list_length (list) == 2)
+                    {
+                      tmpl = g_list_last (list);
+                      if (GTK_IS_COMBO_BOX(tmpl->data))
+                        {
+                          mnemonic_widget = GTK_WIDGET(tmpl->data);
+                        }
+                    }
+                  g_list_free (list);
+                }
+            /*
+             * Handle the case where a GnomeIconEntry is specified as the 
+             * mnemonic widget. use the button which is a grandchild of the
+             * GnomeIconEntry as the mnemonic widget. See bug #133967.
+             */
+              else if (GTK_IS_BOX (mnemonic_widget))
                 {
                   GList *list;
 
@@ -596,6 +639,8 @@ gail_label_get_text (AtkText *text,
     return NULL;
   else
   {
+    if (GAIL_LABEL (text)->textutil == NULL) 
+      gail_label_init_text_util (GAIL_LABEL (text), widget);
     return gail_text_util_get_substring (GAIL_LABEL(text)->textutil, 
                                          start_pos, end_pos);
   }
